@@ -58,9 +58,11 @@ from tempfile import gettempdir
 RETRIES = 3
 MAX_WORKERS = 8
 MAX_ELEMS = 10
+MAX_NESTING = 5
 MEGABYTES = 1024 * 1024
 MAX_BYTES = 2.4 * MEGABYTES
 MAX_PAYLOAD = 15 * MEGABYTES
+MAX_COLUMNS = 100
 DEFAULT_HOST = "contribs-api.materialsproject.org"
 BULMA = "is-narrow is-fullwidth has-background-light"
 PROVIDERS = {"github", "google", "facebook", "microsoft", "amazon"}
@@ -435,6 +437,7 @@ def _load(protocol, host, headers_json, project):
     origin_url = f"{url}/apispec.json"
     fn = urlsafe_b64encode(origin_url.encode('utf-8')).decode('utf-8')
     apispec = Path(gettempdir()) / fn
+    is_mock_test = 'unittest' in sys.modules and protocol == "http"
     spec_dict = None
 
     if apispec.exists():
@@ -444,25 +447,27 @@ def _load(protocol, host, headers_json, project):
         retries, max_retries = 0, 3
         while retries < max_retries:
             try:
-                is_mock_test = 'unittest' in sys.modules and protocol == "http"
-                if is_mock_test or requests.options(f"{url}/healthcheck").status_code == 200:
-                    loader = Loader(http_client)
-                    spec_dict = loader.load_spec(origin_url)
+                if not is_mock_test:
+                    r = requests.options(f"{url}/healthcheck")
+                    if r.status_code != 200:
+                        retries += 1
+                        logger.warning(
+                            f"Healthcheck for {url} failed (Status {r.status_code})! Waiting 60s."
+                        )
+                        time.sleep(60)
+                        continue
 
-                    with apispec.open("w") as f:
-                        ujson.dump(spec_dict, f)
+                loader = Loader(http_client)
+                spec_dict = loader.load_spec(origin_url)
 
-                    logger.debug(f"Specs for {origin_url} saved as {apispec}.")
-                    break
-                else:
-                    retries += 1
-                    logger.warning(
-                        f"Specs not loaded: Healthcheck for {url} failed! Waiting 60s ..."
-                    )
-                    time.sleep(60)
-            except RequestException:
+                with apispec.open("w") as f:
+                    ujson.dump(spec_dict, f)
+
+                logger.debug(f"Specs for {origin_url} saved as {apispec}.")
+                break
+            except RequestException as ex:
                 retries += 1
-                logger.warning(f"Specs not loaded: Could not connect to {url}! Waiting 60s ...")
+                logger.warning(f"Could not connect to {url} ({ex})! Waiting 60s ...")
                 time.sleep(60)
 
     if not spec_dict:
@@ -626,9 +631,17 @@ class Client(SwaggerClient):
         try:
             validate_object(self.swagger_spec, model_spec, data)
         except ValidationError as ex:
-            return str(ex)
+            return False, str(ex)
 
-        return True
+        return True, None
+
+    def _is_serializable_dict(self, dct):
+        for k, v in flatten(dct, reducer="dot").items():
+            if v is not None and not isinstance(v, (str, int, float)):
+                error = f"Value {v} of {type(v)} for key {k} not supported."
+                return False, error
+
+        return True, None
 
     def _get_per_page_default_max(self, op: str = "get", resource: str = "contributions") -> int:
         resource = self.swagger_spec.resources[resource]
@@ -898,7 +911,14 @@ class Client(SwaggerClient):
         if not self.project:
             return {"error": "initialize client with project argument!"}
 
-        columns = columns or {}
+        columns = flatten(columns, reducer="dot") or {}
+
+        if len(columns) > MAX_COLUMNS:
+            return {"error": f"Number of columns larger than {MAX_COLUMNS}!"}
+
+        if not all(isinstance(v, str) for v in columns.values() if v is not None):
+            return {"error": "All values in `columns` need to be None or of type str!"}
+
         new_columns = []
 
         if columns:
@@ -911,12 +931,12 @@ class Client(SwaggerClient):
                     continue
 
                 nesting = k.count(".")
-                if nesting > 4:
-                    return {"error": f"Nesting too deep for {k}"}
+                if nesting > MAX_NESTING:
+                    return {"error": f"Nesting depth larger than {MAX_NESTING} for {k}!"}
 
                 for col in scanned_columns:
                     if nesting and col.startswith(k):
-                        return {"error": f"duplicate definition of {k} in {col}!"}
+                        return {"error": f"Duplicate definition of {k} in {col}!"}
 
                     for n in range(1, nesting+1):
                         if k.rsplit(".", n)[0] == col:
@@ -976,9 +996,9 @@ class Client(SwaggerClient):
                 new_columns.append(new_column)
 
         payload = {"columns": new_columns}
-        valid = self._is_valid_payload("Project", payload)
+        valid, error = self._is_valid_payload("Project", payload)
         if not valid:
-            return {"error": valid}
+            return {"error": error}
 
         return self.projects.update_entry(pk=self.project, project=payload).result()
 
@@ -1326,9 +1346,13 @@ class Client(SwaggerClient):
             return "Nothing to update."
 
         tic = time.perf_counter()
-        valid = self._is_valid_payload("Contribution", data)
+        valid, error = self._is_valid_payload("Contribution", data)
         if not valid:
-            return {"error": valid}
+            return {"error": error}
+
+        serializable, error = self._is_serializable_dict(data)
+        if not serializable:
+            return {"error": error}
 
         query = query or {}
 
@@ -1556,6 +1580,10 @@ class Client(SwaggerClient):
         fields.remove("needs_build")  # internal field
 
         for contrib in tqdm(contributions, desc="Prepare"):
+            serializable, error = self._is_serializable_dict(contrib)
+            if not serializable:
+                raise ValueError(error)
+
             update = "id" in contrib
             project_name = id2project[contrib["id"]] if update else contrib["project"]
             if (
@@ -1673,9 +1701,9 @@ class Client(SwaggerClient):
                     digests[project_name][component].add(digest)
                     contribs[project_name][-1][component].append(dct)
 
-                valid = self._is_valid_payload("Contribution", contribs[project_name][-1])
+                valid, error = self._is_valid_payload("Contribution", contribs[project_name][-1])
                 if not valid:
-                    return {"error": f"{contrib['identifier']} invalid: {valid}!"}
+                    return {"error": f"{contrib['identifier']} invalid: {error}!"}
 
         # submit contributions
         if contribs:
